import urllib.request
import warnings
from pathlib import Path
from typing import Optional

import toolz as tz
from napari.utils import progress

SAM_WEIGHTS_URL = {
    "default": "https://dl.fbaipublicfiles.com/segment_anything/sam_vit_h_4b8939.pth",
    "vit_h": "https://dl.fbaipublicfiles.com/segment_anything/sam_vit_h_4b8939.pth",
    "vit_l": "https://dl.fbaipublicfiles.com/segment_anything/sam_vit_l_0b3195.pth",
    "vit_b": "https://dl.fbaipublicfiles.com/segment_anything/sam_vit_b_01ec64.pth",
}


@tz.curry
def _report_hook(
    block_num: int,
    block_size: int,
    total_size: int,
    pbr: "progress" = None,
) -> None:
    downloaded = block_num * block_size
    percent = downloaded * 100 / total_size
    downloaded_mb = downloaded / 1024 / 1024
    total_size_mb = total_size / 1024 / 1024
    increment = int(percent) - pbr.n
    if increment > 1:  # faster than increment at every iteration
        pbr.update(increment)
    print(
        f"Download progress: {percent:.1f}% ({downloaded_mb:.1f}/{total_size_mb:.1f} MB)",
        end="\r",
    )


def download_weights(weight_url: str, weight_path: "Path"):
    print(f"Downloading {weight_url} to {weight_path} ...")
    pbr = progress(total=100)
    try:
        urllib.request.urlretrieve(
            weight_url, weight_path, reporthook=_report_hook(pbr=pbr)
        )
    except (
        urllib.error.HTTPError,
        urllib.error.URLError,
        urllib.error.ContentTooShortError,
    ) as e:
        warnings.warn(f"Error downloading {weight_url}: {e}")
        return None
    else:
        print("\rDownload complete.                            ")
    pbr.close()


def get_weights_path(model_type: str) -> Optional[Path]:
    """Returns the path to the weight of a given model architecture."""
    weight_url = SAM_WEIGHTS_URL[model_type]

    cache_dir = Path.home() / ".cache/napari-segment-anything"
    cache_dir.mkdir(parents=True, exist_ok=True)

    weight_path = cache_dir / weight_url.split("/")[-1]

    # Download the weights if they don't exist
    if not weight_path.exists():
        download_weights(weight_url, weight_path)

    return weight_path



from typing import Callable

import napari
import numpy as np
import pytest
from numpy.typing import ArrayLike
from skimage.data import astronaut, camera

from napari_segment_anything import SAMWidget


@pytest.fixture
def sample_image() -> ArrayLike:
    return astronaut()


@pytest.mark.parametrize("im_dtype", [np.uint8, np.float32])
def test_click(
    make_napari_viewer: Callable[[], napari.Viewer],
    sample_image: ArrayLike,
    im_dtype: np.dtype,
) -> None:
    viewer = make_napari_viewer()
    # viewer = napari.Viewer()
    sample_image = sample_image.astype(im_dtype)

    widget = SAMWidget(viewer, model_type="vit_b")

    viewer.window.add_dock_widget(widget)
    viewer.add_image(sample_image)

    assert widget._predictor is not None
    assert widget._im_layer_widget.value is not None
    assert not widget._confirm_mask_btn.enabled
    assert not widget._cancel_annot_btn.enabled

    # first interaction
    widget._pts_layer.data = [[42, 233]]  # point on hair

    assert np.any(widget._mask_layer.data > 0)
    assert np.all(widget._labels_layer.data == 0)
    assert widget._confirm_mask_btn.enabled
    assert widget._cancel_annot_btn.enabled

    # confirm mask
    widget._confirm_mask_btn.clicked()
    assert np.all(widget._mask_layer.data == 0)
    assert np.any(widget._labels_layer.data > 0)
    assert not widget._confirm_mask_btn.enabled
    assert not widget._cancel_annot_btn.enabled

    # new interaction
    widget._pts_layer.data = [[42, 233], [125, 225]]  # adding point to face
    assert np.any(widget._mask_layer.data > 0)

    # confirm mask
    widget._confirm_mask_btn.clicked()
    assert np.all(widget._mask_layer.data == 0)
    assert len(np.unique(widget._labels_layer.data)) == 3

    # test cancel interaction
    widget._pts_layer.data = [[42, 233]]
    assert np.any(widget._mask_layer.data > 0)

    widget._cancel_annot_btn.clicked()
    assert np.any(widget._mask_layer.data == 0)
    assert len(np.unique(widget._labels_layer.data)) == 3  # still the same
    assert not widget._confirm_mask_btn.enabled
    assert not widget._cancel_annot_btn.enabled


@pytest.mark.parametrize("image", [astronaut(), camera()])
def test_automatic_segmentation(
    make_napari_viewer: Callable[[], napari.Viewer],
    image: ArrayLike,
) -> None:
    viewer = make_napari_viewer()
    widget = SAMWidget(viewer, model_type="vit_b")

    viewer.window.add_dock_widget(widget)
    viewer.add_image(image)

    widget._auto_segm_btn.clicked()
    assert np.any(widget._labels_layer.data > 0)
